# History tab for NeuroLab
# Martin Reiche 2024
import tkinter as tk
from collections import defaultdict
import Global as g

class History(tk.Frame):
        
    def prepare_history(self):
        history = len(g.history)
        if history > 0:
            self.prepare_records()
            self.prepare_errors()
            self.prepare_losses()
            self.create_kohonen_display()
            self.show_winning_neurons(self.last_epoch)
            self.show_error(self.last_epoch)
            self.show_loss(self.last_epoch)
            self.current_epoch.set(self.last_epoch)
            self.spin_box_epoch.selection_clear()
            self.spin_box_epoch.icursor(99)


    def prepare_records(self):  # collect winnig neuron entries in history
        self.records = {}       # key: epoch, value: list of label-place tuple
        last_epoch = 1
        epochs = [last_epoch]
        winners = []
        for epoch, output_index, winning_neuron in g.history:
            if epoch == last_epoch:
                winners.append((winning_neuron, output_index))
            else:
                self.records[last_epoch] = winners
                winners = [(winning_neuron, output_index)] # start new list
                epochs.append(epoch)
                last_epoch = epoch
                
        self.records[last_epoch] = winners
        self.last_epoch = last_epoch
        self.spin_box_epoch.configure(values=epochs)

    def prepare_errors(self):
        self.errors = {}
        for epoch, error in g.align_errors:
            self.errors[epoch] = error

    def prepare_losses(self):
        self.losses = {}
        for epoch, loss in g.proximity_losses:
            self.losses[epoch] = loss
        
    def show_winning_neurons(self, epoch):
        self.canvas_kohonen.delete('position')
        data = self.records[epoch]
        self.assignments = defaultdict(list) # key: position, value: list of labels
        # print('epoch', epoch)
        for (row, col), pattern_index in data:
            label = g.training_pattern_labels[pattern_index]
            self.assignments[(row, col)].append(label)
            # print('label', label, 'in row', row, 'col', col)
            x = self.kohonen_neuron_size*(col + 0.5) + self.border
            y = self.kohonen_neuron_size*(row + 0.5)
            self.canvas_kohonen.create_text(x, y, font=g.fontLabel, text=label,
                                            tags='position')
    def show_error(self, epoch):
        error = self.errors[epoch]
        self.current_error.set(f'{error:0.2f}')

    def show_loss(self, epoch):
        loss = self.losses[epoch]
        self.current_loss.set(loss)
    
    def spin_epoch(self):
        epoch = self.current_epoch.get()
        self.show_winning_neurons(epoch)
        self.show_error(epoch)
        self.show_loss(epoch)

    def spin_box_epoch_changed(self, _):
        epoch = self.current_epoch.get()
        if self.records.get(epoch):
            self.show_winning_neurons(epoch)
            self.show_error(epoch)
            self.show_loss(epoch)
        
    def create_kohonen_display(self):
        number_rows = number_cols = g.kohonen_layer_size
        self.kohonen_neuron_size = min(self.kohonen_side // number_rows,
                                       self.kohonen_side // number_cols)
        if self.kohonen_neuron_size > 40: self.kohonen_neuron_size = 40
        for row in range(number_rows):
            for col in range(number_cols):
                x0 = self.kohonen_neuron_size * col + self.border
                x1 = x0 + self.kohonen_neuron_size
                y0 = self.kohonen_neuron_size * row
                y1 = y0 + self.kohonen_neuron_size
                handle = self.canvas_kohonen.create_rectangle(x0, y0, x1, y1,
                                                           fill=g.bgLight,
                                                           outline="#808080")

    def kohonen_mouse_hover(self, event):
        if not self.kohonen_neuron_size:
            return
        row = event.y // self.kohonen_neuron_size
        if row >= g.kohonen_layer_size: return
        col = (event.x - self.border) // self.kohonen_neuron_size
        if col >= g.kohonen_layer_size: return
        
        if labels := self.assignments.get((row, col)):
            self.text_assigned_labels.config(state=tk.NORMAL)
            self.text_assigned_labels.delete(1.0, tk.END)
            self.text_assigned_labels.insert(tk.END, labels)
            self.text_assigned_labels.config(state=tk.DISABLED)
        else:
            self.text_assigned_labels.config(state=tk.NORMAL)
            self.text_assigned_labels.delete(1.0, tk.END)
            self.text_assigned_labels.config(state=tk.DISABLED)

    def mousewheel_epoch(self, event):
        if event.delta > 0:
            self.spin_box_epoch.invoke('buttonup')
        else:
            self.spin_box_epoch.invoke('buttondown')
        
    lineSpace = 40
    header_line = 20
    row_1 = 60
    kohonen_side = 360
    border = 20
    column_1 = 20
    column_2 = column_1 + 120
    column_3 = column_2 + kohonen_side + 2 * border + 10
    column_4 = column_3 + 120
    border = 20        
        
    def __init__(self, notebook):
        super(History, self).__init__(notebook, background=g.bgDark)
        
        rows = [self.row_1 + i * self.lineSpace for i in range(10)]
        self.kohonen_neuron_size = None
        
        # Left Column ==============================================================
        lbl = tk.Label(self, text='Epoch', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.header_line)
        
        self.current_epoch = tk.IntVar()
        self.current_epoch.set(0)
        self.spin_box_epoch = tk.Spinbox(self, from_=1, to=1, textvariable=self.current_epoch,
                                          font=g.fontTitle, width=4, background=g.bgDark, command=self.spin_epoch)
        self.spin_box_epoch.place(x=self.column_1, y=rows[0] + 2)
        self.spin_box_epoch.bind('<Return>', self.spin_box_epoch_changed)
        self.spin_box_epoch.bind('<MouseWheel>', self.mousewheel_epoch)
        
        # Center Column =============================================================
        lbl = tk.Label(self, text='Winning neurons (BMUs)',
                       font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_2, y=self.header_line)
        
        width = self.kohonen_side + 2 * self.border
        height = self.kohonen_side
        self.canvas_kohonen = tk.Canvas(self, width=width, height=height,
                              bg=g.bgDark, highlightbackground=g.bgDark)
        self.canvas_kohonen.place(x=self.column_2 - self.border, y=rows[0])
        self.canvas_kohonen.bind('<Motion>', self.kohonen_mouse_hover)
        
        # Right Column =============================================================
        self.current_error = tk.StringVar()
        lbl = tk.Label(self, text='Error', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=rows[0])
        lbl = tk.Label(self, textvariable=self.current_error,
                           font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=rows[0])

        self.current_loss = tk.IntVar()
        lbl = tk.Label(self, text='Proximity loss', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=rows[1])
        lbl = tk.Label(self, textvariable=self.current_loss,
                           font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=rows[1])
        
        # Assigned Labels
        text = 'Labels assigned to\nselected neuron:'
        lbl = tk.Label(self, text=text, font=g.fontLabel, justify=tk.LEFT, background=g.bgDark)
        lbl.place(x=self.column_3, y=rows[2])

        self.text_assigned_labels = \
            tk.Text(self, font=g.fontLabel, height=13, width=20, background=g.bgLight,
                                    relief=tk.FLAT, state=tk.DISABLED,  wrap=tk.WORD)
        self.text_assigned_labels.place(x=self.column_3 + 4, y=rows[3] + 10)