# Setup, train & test tab for NeuroLab
# Martin Reiche 2024
import tkinter as tk
from tkinter import filedialog
import time
import numpy as np
import matplotlib.pyplot as plt
import json
import random
import copy
import Global as g
import FileAccess
from KohonenNet import KohonenNet

class SetTrainTest(tk.Frame):

    def closest_pairs_loss(self, patterns, number_closest):
        num_patterns = len(patterns)
        pattern_length = len(patterns[0][1])
        close_pairs = []
        for i in range(num_patterns):
            for j in range(i+1, num_patterns):
                # Determine Hamming Distance
                dist = 0
                for k in range(pattern_length):
                    if g.all_training_patterns[i][1][k] != \
                       g.all_training_patterns[j][1][k]:
                       dist += 1
                close_pairs.append(((i, j), dist))
        close_pairs.sort(key=lambda x: x[1])
        
#         for i, pair in enumerate(close_pairs):
#             print(str(i+1)+':', g.training_pattern_labels[pair[0][0]],'-', \
#                    g.training_pattern_labels[pair[0][1]], ':', pair[1])
    
        max_dist = close_pairs[number_closest-1][1]
        close_pairs = close_pairs[:number_closest]
        # Create loss penalty
        for m in range(number_closest):
            (i, j), dist = close_pairs[m]
            penalty = max_dist - dist
            close_pairs[m] = (i, j), penalty

        return close_pairs

    def calculate_epoch_proximity_loss(self, epoch_results):
        loss = 0
        for (index_1, index_2), penalty in g.proximity_penalty:
            for loc, output_index in epoch_results:
                if output_index == index_1:
                    loc_1 = loc
                    break 
            for loc, output_index in epoch_results:
                if output_index == index_2:
                    loc_2 = loc
                    break
            if loc_1 == loc_2: # both labels in same place
                loss += 10 * g.proximity_penalty[0][1] # 10 is an arbitrary choice
            else: # calculate Manhattan distance
                dist = abs(loc_1[0] - loc_2[0]) + abs(loc_1[1] - loc_2[1])
                loss += (dist - 1) * penalty
        return loss
        
        
    def calculate_proximity_loss_curve(self):
        if len(g.history) == 0:
            return
        g.proximity_losses = []
        last_epoch = 1
        epoch_results = []
        for epoch, output_index, winning_neuron in g.history:
            if epoch == last_epoch:
                epoch_results.append((winning_neuron, output_index))
            else:
                prox_loss = self.calculate_epoch_proximity_loss(epoch_results)
                g.proximity_losses.append((last_epoch, prox_loss))
                epoch_results = [(winning_neuron, output_index)] # start new list
                last_epoch = epoch
        prox_loss = self.calculate_epoch_proximity_loss(epoch_results)
        g.proximity_losses.append((last_epoch, prox_loss))            
                
    def load_training_data(self):
        files = tk.filedialog.askopenfilenames()
        if len(files) == 0:  # User cancelled
            return

        # clear test data
        g.all_test_patterns = []
        g.test_pattern_labels = []
        self.number_test_records.set('')
        g.test_file_name.set('')
        
        g.training_file_name.set("Training data: " + files[0])
        g.gui.update_status('Loading training data, please wait!')
        success, num_rows, num_cols, encoding = \
                 FileAccess.load_data_file(files[0], g.all_training_patterns,
                                           g.training_pattern_labels)
        
        g.gui.update_status('')
        if success:
            if g.num_input_layer_rows.get():
                if num_rows != int(g.num_input_layer_rows.get()) or \
                    num_cols != int(g.num_input_layer_cols.get()):
                    g.kohonen_net = None
            g.num_input_layer_rows.set(num_rows)
            g.num_input_layer_cols.set(num_cols)
            self.number_training_records.set(len(g.all_training_patterns))
            
        if encoding == 'data_bw':
            n = 2 * len(g.all_training_patterns) # number of proximity relations
            g.proximity_penalty = self.closest_pairs_loss(g.all_training_patterns, n)
            
          
    def load_test_data(self):
        if len(g.all_training_patterns) == 0:
            msg = 'Please load training data first!'
            tk.messagebox.showerror('Cannot load test data', msg)
            return
        files = tk.filedialog.askopenfilenames()
        if len(files) == 0:  # User cancelled
            return

        g.test_file_name.set("Test data: " + files[0])
        g.gui.update_status('Loading test data, please wait!')
        success, num_rows, num_cols, encoding = \
                 FileAccess.load_data_file(files[0], g.all_test_patterns,
                                           g.test_pattern_labels)
        g.gui.update_status('')
        if success:
            if num_rows != int(g.num_input_layer_rows.get()) or \
               num_cols != int(g.num_input_layer_cols.get()):
                msg = 'Test data shape doesn\'t match training data'
                tk.messagebox.showerror('Inappropriate test data', msg)
                g.all_test_patterns = []
                return
            self.number_test_records.set(len(g.all_test_patterns))
    
    def run_training(self):
        if len(g.all_training_patterns) == 0:
            msg = 'Please load training data!'
            tk.messagebox.showerror('Cannot run training', msg)
            return
        g.gui.update_status('Training in progress, please wait!')
        # Check parameter
        try:
            kohonen_layer_size = int(self.kohonen_size.get())
            if g.kohonen_net:
                if kohonen_layer_size != g.kohonen_layer_size:
                    g.kohonen_net = None
            if kohonen_layer_size < 1 or kohonen_layer_size > 50:
                raise ValueError
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Kohonen layer size has an invalid value')
            return

        try:
            learning_rate_begin = float(self.learning_rate_begin.get())
            if learning_rate_begin <= 0 or learning_rate_begin > 100:
                raise ValueError
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Learning rate begin has an invalid value')
            return

        try:
            learning_rate_end = float(self.learning_rate_end.get())
            if learning_rate_end <= 0 or learning_rate_end > 100:
                raise ValueError
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Learning rate end has an invalid value')
            return

        try:
            radius_begin = float(self.radius_begin.get())
            if radius_begin <= 0 or radius_begin > 100:
                raise ValueError
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Radius begin has an invalid value')
            return

        try:
            radius_end = float(self.radius_end.get())
            if radius_end <= 0 or radius_end > 100:
                raise ValueError
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Rsdius end has an invalid value')
            return

        try:
            epochs = int(self.epochs.get())
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Epochs has an invalid value')
            return

        try:
            random_seed = int(self.random_seed.get())
        except ValueError:
            tk.messagebox.showinfo('Setup error:', 'Random seed has an invalid value')
            return
               
        if not g.kohonen_net or kohonen_layer_size != g.kohonen_layer_size:
            g.kohonen_layer_size = kohonen_layer_size
            if not g.num_input_layer_rows.get():
                return

            num_input_rows = int(g.num_input_layer_rows.get())
            num_input_cols = int(g.num_input_layer_cols.get())

            g.kohonen_net = KohonenNet(num_input_rows,
                                       num_input_cols,
                                       g.kohonen_layer_size,
                                       g.kohonen_layer_size,
                                       random_seed)
            
        g.tab_kohonen_map.create_display()
        start_time = time.perf_counter()
        self.training_time.set('')
        self.last_error.set('')
        self.current_epoch.set('')
        g.align_errors = []
        g.proximity_losses = []
        g.history = []
        learning_rate = learning_rate_begin
        radius = radius_begin
        if epochs > 1:
            delta_learning_rate = (learning_rate_end - learning_rate_begin) / (epochs - 1)
            delta_radius = (radius_end - radius_begin) / (epochs - 1)
        else:
            delta_learning_rate = 0
            delta_radius = 0
        training_patterns = copy.copy(g.all_training_patterns)
        e = error = 0.0
        for e in range(epochs):
            random.shuffle(training_patterns)
            if e % 10 == 9:     # display update only every 10 epochs (performance)
                self.current_epoch.set(str(e+1))
                elapsed = time.perf_counter() - start_time
                self.training_time.set(int(elapsed))
                self.update()
            if e == 0 or e % 10 == 9: # Capture error info every 10 epochs and last
                epoch = e + 1
            else:
                epoch = None
            error = g.kohonen_net.train(epoch, training_patterns, learning_rate, radius)
            g.align_errors.append((e+1, error))
            learning_rate += delta_learning_rate
            radius += delta_radius
        g.align_errors.append((e, error))
        self.last_error.set(f'{error:0.2f}')
        self.calculate_proximity_loss_curve()
        g.gui.update_status('')   

    def show_error_curve(self):
        plt.close()
        x = []; y = []
        for epoch, error in g.align_errors:
            x.append(epoch)
            y.append(error)
        plt.plot(x, y)
        plt.xlabel('Epoch')
        plt.ylabel('Error')
        plt.title('Specialization of Winning Neurons (BMUs)')
        plt.show()

    def show_proximity_loss_curve(self):
        plt.close()
        x = []; y = []
        for epoch, loss in g.proximity_losses:
            x.append(epoch)
            y.append(loss)
        plt.plot(x, y)
        plt.xlabel('Epoch')
        plt.ylabel('Proximity loss')
        plt.title('Distribution of Winning Neurons (BMUs)')
        plt.show()
        
    def reset_kohonen(self):
        g.kohonen_net = None
        self.current_epoch.set('')
        self.training_time.set('')
        self.last_error.set('')
    
    def save_kohonen(self):
        if g.kohonen_net:
            data = g.kohonen_net.dump()
            outfile = open('kohonen_layer.txt', 'w')
            json.dump(data, outfile, indent=2)
            outfile.close()
            
    column_1 = 20
    column_2 = column_1 + 150
    column_3 = column_2 + 110
    column_4 = column_3 + 150
    column_5 = column_4 + 110
    column_6 = column_5 + 140
    column_7 = column_6 + 60

    lineSpace = 40
    header_line = 20
    row_1 = 60
            
    def __init__(self, notebook):
        super(SetTrainTest, self).__init__(notebook, background=g.bgDark)
        
        self.state = None
        self.callback = None
        self.number_training_records = tk.StringVar()
        self.number_test_records = tk.StringVar()
        self.kohonen_size        = tk.StringVar()
        self.kohonen_size.set('10')
        self.learning_rate_begin = tk.StringVar()
        self.learning_rate_begin.set('0.9')
        self.learning_rate_end   = tk.StringVar()
        self.learning_rate_end.set('0.1')
        self.radius_begin        = tk.StringVar()
        self.radius_begin.set('4.0')
        self.radius_end        = tk.StringVar()
        self.radius_end.set('0.1')
        self.epochs              = tk.StringVar()
        self.epochs.set("500")
        self.random_seed         = tk.StringVar()
        self.random_seed.set("1")
        self.current_epoch       = tk.StringVar()
        self.training_time       = tk.StringVar()
        self.last_error          = tk.StringVar()
        
        g.num_input_layer_rows = tk.StringVar()
        g.num_input_layer_cols = tk.StringVar()
        
        self.rows = [self.row_1 + n * self.lineSpace for n in range(13)]
        
        # Left Column ==============================================================
        lbl = tk.Label(self, text='Setup', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.header_line)

        # Load trainig data
        button = tk.Button(self, text="Load Training Data", width=21,
                               font=g.fontLabel, background=g.bgDark, command=self.load_training_data)
        button.place(x=self.column_1, y=self.rows[0] - 5)

        # Input Layer Width
        lbl = tk.Label(self, text='Input Layer Width', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[1])
        lbl = tk.Label(self, textvariable=g.num_input_layer_cols,
                                               font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_2, y=self.rows[1])

        # Input Layer Height
        lbl = tk.Label(self, text='Input Layer Height', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[2])
        lbl = tk.Label(self, textvariable=g.num_input_layer_rows,
                                                font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_2, y=self.rows[2])

        # Kohonen layer size
        lbl = tk.Label(self, text='Kohonen Layer Size', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[3])
        entry = tk.Entry(self, textvariable=self.kohonen_size, justify=tk.CENTER,
                         font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[3])

        # Learning rate
        lbl = tk.Label(self, text='Learning\nRate', font=g.fontLabel,
                       justify=tk.LEFT, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[4] + 9)
        
        # Learning rate begin
        lbl = tk.Label(self, text='begin', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1 + 90, y=self.rows[4])
        entry = tk.Entry(self, textvariable=self.learning_rate_begin, justify=tk.CENTER,
                           font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[4])

        # Learning rate end
        lbl = tk.Label(self, text='end', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1 + 90, y=self.rows[5])
        entry = tk.Entry(self, textvariable=self.learning_rate_end, justify=tk.CENTER,
                           font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[5])

        # Radius
        lbl = tk.Label(self, text='Radius', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[6] + 16)

        # Radius begin
        lbl = tk.Label(self, text='begin', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1 + 90, y=self.rows[6])
        entry = tk.Entry(self, textvariable=self.radius_begin, justify=tk.CENTER,
                           font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[6])

        # Radius end
        lbl = tk.Label(self, text='end', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1 + 90, y=self.rows[7])
        entry = tk.Entry(self, textvariable=self.radius_end, justify=tk.CENTER,
                           font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[7])

        # Training Epochs
        lbl = tk.Label(self, text='Training Epochs', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[8])
        entry = tk.Entry(self, textvariable=self.epochs, justify=tk.CENTER,
                            font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[8])

        # Random Seed
        lbl = tk.Label(self, text='Random Seed', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[9])
        entry = tk.Entry(self, textvariable=self.random_seed, justify=tk.CENTER,
                            font=g.fontLabel, width=5, background=g.bgBlue)
        entry.place(x=self.column_2, y=self.rows[9])
        
        # Middle Column ==============================================================
        lbl = tk.Label(self, text='Train', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.header_line)

        # Run training button
        button = tk.Button(self, text="Run Training", width=21,
                               font=g.fontLabel, background=g.bgDark, command=self.run_training)
        button.place(x=self.column_3, y=self.rows[0] - 5)

        # Number of training records
        lbl = tk.Label(self, text='Training Records', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.rows[1])
        lbl = tk.Label(self, textvariable=self.number_training_records,
                                        font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=self.rows[1])

        # Current Epoch
        lbl = tk.Label(self, text='Current Epoch', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.rows[2])
        lbl = tk.Label(self, textvariable=self.current_epoch,
                                           font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=self.rows[2])

        # Training duration
        lbl = tk.Label(self, text='Time Elapsed [sec]', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.rows[3])
        lbl = tk.Label(self, textvariable=self.training_time,
                                           font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=self.rows[3])

        # Last error
        lbl = tk.Label(self, text='Last error', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.rows[4])
        lbl = tk.Label(self, textvariable=self.last_error,
                                           font=g.fontLabel, width=5, background=g.bgLight)
        lbl.place(x=self.column_4, y=self.rows[4])
        
        # Show Error Curve
        button_error_curve = tk.Button(self, text="Show Error Curve", width=21, font=g.fontLabel, background=g.bgDark,
                                       command=self.show_error_curve)
        button_error_curve.place(x=self.column_3, y=self.rows[5])
        
        # Show Proximity loss Curve
        button_error_curve = tk.Button(self, text="Show Proximity Loss Curve", width=21, font=g.fontLabel, background=g.bgDark,
                                       command=self.show_proximity_loss_curve)
        button_error_curve.place(x=self.column_3, y=self.rows[6])

        # Reset button
        button = tk.Button(self, text="Reset Kohonen Layer", width=21,
                                 font=g.fontLabel, background=g.bgDark, command=self.reset_kohonen)
        button.place(x=self.column_3, y=self.rows[7])

        # Store button
        button = tk.Button(self, text="Save Kohonen Layer", width=21,
                                 font=g.fontLabel, background=g.bgDark, command=self.save_kohonen)
        button.place(x=self.column_3, y=self.rows[8])
        
        # Right Column ==============================================================
        lbl = tk.Label(self, text='Test', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_5, y=self.header_line)

        # Load test data
        button_Load_test = tk.Button(self, text="Load Test Data", width=21,
                               font=g.fontLabel, background=g.bgDark, command=self.load_test_data)
        button_Load_test.place(x=self.column_5, y=self.rows[0] - 5)

        # Number of test records
        lbl = tk.Label(self, text='Test Records', font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_5, y=self.rows[1])
        lbl = tk.Label(self, textvariable=self.number_test_records,
                                          font=g.fontLabel, width=6, background=g.bgLight)
        lbl.place(x=self.column_6, y=self.rows[1])
        
        # Hint (use multiple labels as multiline doesn't work)
        text = 'As long as no test data is loaded,\n' \
               'the response to training data can\n' \
               'be viewed in tab \'Kohonen Map\''
        lbl = tk.Label(self, text=text, font=g.fontLabel, justify=tk.LEFT, background=g.bgDark)
        lbl.place(x=self.column_5, y=self.rows[2])
